| package |
package := Package name: 'IDB Duration'.
package paxVersion: 1;
	basicComment: 'Implementation of the ANSI <Duration> protocol
Version 6a - May 2005
For Version 6 of Dolphin Smalltalk

(C) 2005 Ian Bartholomew
ian@idb.me.uk
Public Domain Freeware

Some methods provided by Richard Harmon
'.

package basicPackageVersion: '1'.


package classNames
	add: #Duration;
	add: #DurationToText;
	yourself.

package methodNames
	add: #Float -> #asRestrictedFraction;
	add: #Number -> #asRestrictedFraction;
	add: #Number -> #divideIntoDuration:;
	add: #String -> #trimZeros;
	add: 'Time class' -> #fromDuration:;
	yourself.

package binaryGlobalNames: (Set new
	yourself).

package globalAliases: (Set new
	yourself).

package setPrerequisites: (IdentitySet new
	add: '..\Object Arts\Dolphin\Base\Dolphin';
	add: '..\Object Arts\Dolphin\MVP\Type Converters\Dolphin Type Converters';
	yourself).

package!

"Class Definitions"!

Magnitude subclass: #Duration
	instanceVariableNames: 'seconds'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!
AbstractToTextConverter subclass: #DurationToText
	instanceVariableNames: 'format'
	classVariableNames: ''
	poolDictionaries: ''
	classInstanceVariableNames: ''!

"Global Aliases"!


"Loose Methods"!

!Float methodsFor!

asRestrictedFraction
	"Overridden to restrict the answered Fraction to the equivalent of six decimal
	places in the receiver. This allows any precision lost in the conversion to be ignored"

	#idbAdded.
	^(self fractionPart * 1000000) rounded / 1000000 + self integerPart asInteger! !
!Float categoriesFor: #asRestrictedFraction!converting!idb goodies!public! !

!Number methodsFor!

asRestrictedFraction
	"Answer the receiver as a Fraction.  Subclasses can override to restrict 
	the conversion to a value that is an accurate representetion of the receiver"

	#idbAdded.
	^self asFraction!

divideIntoDuration: aDuration 
	"Private - Answer a new <Duration> which represents the the value in 
	seconds of aDuration divided by the receiver"

	#idbAdded.
	^Duration seconds: aDuration asSeconds / self! !
!Number categoriesFor: #asRestrictedFraction!converting!idb goodies!public! !
!Number categoriesFor: #divideIntoDuration:!double dispatch!idb goodies!public! !

!String methodsFor!

trimZeros
	"Answer the receiver but with trailing 0s removed from the fractional part (if any).
	xxxx.0123 -> xxx.0123
	xxxx.0123000 -> xxxx.0123
	xxxx.0000 -> xxxx.0
	xxxx.0 -> xxxx.0
	xxx0 -> xxx0"

	| index |
	#idbAdded.
	(self includes: $.) ifFalse: [^self copy].
	index := self size.
	[(self at: index) = $0 and: [index > 1 and: [(self at: index - 1) isDigit]]] whileTrue: 
			[self at: index put: Character space.
			index := index - 1].
	^self trimBlanks! !
!String categoriesFor: #trimZeros!copying!idb goodies!public! !

!Time class methodsFor!

fromDuration: aDuration 
	"Answer an instance of Time initialized from aDuration.  
	NB: 
		Any day value in aDuration is ignored
		Negative values for aDuration produce backward Times i.e.
			Time fromDuration: (Duration days: 0 hours: -1 minutes: -2 seconds: -3) answers 22:57:57"

	#idbAdded.
	^self 
		fromMilliseconds: (((aDuration hours * 60 + aDuration minutes) * 60 + aDuration seconds) * 1000) 
				rounded! !
!Time class categoriesFor: #fromDuration:!idb goodies!instance creation!public! !

"End of package definition"!

"Source Globals"!

"Classes"!

Duration guid: (GUID fromString: '{87B4C654-026E-11D3-9FD7-00A0CC3E4A32}')!
Duration comment: 'See [DolphinImageFolder]/Idb/Documentation for details

(C) 2005 Ian Bartholomew
ian@idb.me.uk
Public Domain Freeware'!
!Duration categoriesForClass!IDB Goodies!Magnitude-General! !
!Duration methodsFor!

- operand 
	"Answer a <Duration> whose value is the result of 
	subtracting the <Duration> operand from the receiver"

	^self class seconds: seconds - operand asSeconds!

* operand 
	"Answer a <Duration> whose value is the result of 
	multiplying the <number> operand by the receiver"

	^self class seconds: seconds * operand!

/ operand 
	"If operand is a <number> answer a new <Duration> whose value is the result of 
	dividing the receiver by operand.  If operand is a <Duration> answer a <number> 
	whose value is the result of dividing  the receiver by operand. 	If operand equals zero
	or <Duration factory> #zero the ZeroDivide exception is signaled"

	^operand divideIntoDuration: self!

+ operand 
	"Answer a <Duration> whose value is the result of 
	adding the <Duration> operand to the receiver"

	^self class seconds: seconds + operand asSeconds!

< operand 
	"Answer true if operand represents a <Duration> that is larger than the receiver"

	^seconds < operand asSeconds!

= comparand 
	"Answer true if the comparand is a <Duration> representing the 
	same length of time as the receiver"

	^self species == comparand species and: [seconds = comparand asSeconds]!

> operand 
	"Answer true if operand represents a <Duration> that is smaller than the receiver"

	^seconds > operand asSeconds!

abs
	"If the receiver is greater than or equal to <Duration Factory> #zero answer 
	a <Duration> which is equal to the receiver. Otherwise answer a <Duration> 
	which has the same magnitude as the receiver but the opposite sign"

	^self class seconds: seconds abs!

asSeconds
	"Answer the total number of seconds in the length of time represented by the receiver
	including any fractional part of a second.  If the receiver is less than <Duration factory>
	#zero then the result will be less than 0
	(idb) Always answers a Fraction or an Integer"

	^seconds!

days
	"Answer the the <integer> number of whole days represented by the receiver.
	The answer may be negative for negative durations.
	(idb) see comment in #hours"

	^seconds quo: ##(24 * 60 * 60)!

divideIntoDuration: aDuration 
	"Part of a double dispatch for <Duration> division.  Answers a <number>"

	^aDuration asSeconds / seconds!

hash
	"Answer the SmallInteger hash value for the receiver."

	^seconds hash!

hours
	"Answer an <integer> beteen -23 and 23, inclusive, which is the number 
	of whole hours represented by the receiver after complete days have been 
	removed.
	(idb) Use rem and quo rather than \\ and // to preserve sign on negative values
	-7200 \\ ##(24 * 60 * 60) --> 79200 ... wrong
	-7200 rem: ##(24 * 60 * 60) --> -7200

	(-7201 rem: ##(24 * 60 * 60)) // ##(60 * 60) --> -3 ... wrong
	(-7201 rem: ##(24 * 60 * 60)) quo: ##(60 * 60) --> -2"

	^(seconds rem: ##(24 * 60 * 60)) quo: ##(60 * 60)!

microseconds
	"Answer the number of microseconds in the fractional part of the receiver"

	^(seconds fractionPart * 1000000) rounded!

milliseconds
	"Answer the number of milliseconds in the fractional part of the receiver"

	^(seconds fractionPart * 1000) rounded!

minutes
	"Answer an <integer> beteen -59 and 59, inclusive, which is the number 
	of whole minutes represented by the receiver after complete days and hours
	have been removed.
	(idb) see comment in #hours"

	^(seconds rem: ##(60 * 60)) quo: 60!

negated
	"Answer a <Duration> of the opposite sign as the receiver."

	^self class seconds: seconds negated!

negative
	"Answer whether the receiver is negative."

	^seconds < 0!

positive
	"Answer whether the receiver is positive or zero."

	^seconds >= 0!

printOn: target 
	"Append a default string representation of the receiver to the <readableStream>,
	target. The result will be of the format: [-]D:HH:MM:SS[.S]
	(idb) Any fractional part is displayed as a Float with a maximum of 6 decimal places"

	self negative ifTrue: [target nextPut: $-].
	target
		print: self days abs;
		nextPutAll: (':%02d' sprintfWith: self hours abs);
		nextPutAll: (':%02d' sprintfWith: self minutes abs).
	target nextPut: $:.
	self seconds abs < 10 ifTrue: [target nextPut: $0].
	self seconds fractionPart isZero 
		ifTrue: [target print: self seconds abs]
		ifFalse: 
			[| s |
			s := String writeStream.
			self seconds abs asFloat printOn: s decimalPlaces: 6.
			target nextPutAll: s contents trimZeros]!

seconds
	"Answer a <number> in the range -60 < s < 60, which is the number 
	of seconds, including any fraction part, represented by the receiver after 
	complete days, hours and minutes have been removed.
	(idb) see comment in #hours"

	^seconds rem: 60!

setSeconds: aNumber 
	"Set the <number> of seconds that the receiver represents
	(idb) This will always be an Integer or Fraction - never a Float"

	seconds := aNumber! !
!Duration categoriesFor: #-!arithmetic!public! !
!Duration categoriesFor: #*!arithmetic!public! !
!Duration categoriesFor: #/!arithmetic!public! !
!Duration categoriesFor: #+!arithmetic!public! !
!Duration categoriesFor: #<!comparing!public! !
!Duration categoriesFor: #=!comparing!public! !
!Duration categoriesFor: #>!comparing!public! !
!Duration categoriesFor: #abs!arithmetic!public! !
!Duration categoriesFor: #asSeconds!converting!public! !
!Duration categoriesFor: #days!accessing!public! !
!Duration categoriesFor: #divideIntoDuration:!arithmetic!double dispatch!public! !
!Duration categoriesFor: #hash!comparing!public! !
!Duration categoriesFor: #hours!accessing!public! !
!Duration categoriesFor: #microseconds!accessing!public! !
!Duration categoriesFor: #milliseconds!accessing!public! !
!Duration categoriesFor: #minutes!accessing!public! !
!Duration categoriesFor: #negated!arithmetic!public! !
!Duration categoriesFor: #negative!public!testing! !
!Duration categoriesFor: #positive!public!testing! !
!Duration categoriesFor: #printOn:!printing!public! !
!Duration categoriesFor: #seconds!accessing!public! !
!Duration categoriesFor: #setSeconds:!instance creation!public! !

Duration methodProtocol: #Duration attributes: #(#ansi #readOnly) selectors: #(#- #* #/ #~~ #~= #+ #< #<= #= #== #> #>= #abs #asSeconds #between:and: #class #copy #days #doesNotUnderstand: #error: #hash #hours #identityHash #isKindOf: #isMemberOf: #isNil #max: #min: #minutes #negated #negative #notNil #perform: #perform:with: #perform:with:with: #perform:with:with:with: #perform:withArguments: #positive #printOn: #printString #respondsTo: #seconds #yourself)!

!Duration class methodsFor!

days: days 
	"Answer a <Duration> representing days. If days is negative then
	answer a <Duration> less than <Duration factory>#zero"

	^self 
		days: days
		hours: 0
		minutes: 0
		seconds: 0!

days: days hours: hours minutes: minutes seconds: seconds 
	"Answer a <Duration> of the number of days, hours, minutes, and seconds. If any 
	of the operands are negative, the result is smaller by that number of days, hours, 
	minutes, or seconds as appropriate."

	^self seconds: ((days * 24 + hours) * 60 + minutes) * 60 + seconds!

errorInvalidFormat
	"Private - Raise an error that some attempt to create an instance of the receiver failed because the some input was in the wrong format."

	#DtTmX.	"2001/07/18 Harmon, R. Added for DateAndTimeExtra support."
	^InvalidFormat signalWith: self!

fromString: aString 
	"Answer an instance of the receiver constructed from aString <String> in the Reg. Ex. format: -*D:HH:MM:SS(.SS*)?."

	#DtTmX.	"2001/07/18 Harmon, R. Added for <DurationExtra factory>."
	^self readFrom: (ReadStream on: aString)!

hours: hours 
	"Answer a <Duration> representing hours. If hours is negative then
	answer a <Duration> less than <Duration factory>#zero"

	^self 
		days: 0
		hours: hours
		minutes: 0
		seconds: 0!

minutes: minutes 
	"Answer a <Duration> representing minutes. If minutes is negative then
	answer a <Duration> less than <Duration factory>#zero"

	^self 
		days: 0
		hours: 0
		minutes: minutes
		seconds: 0!

readFrom: aStream 
	"Answers a new instance of the receiver read from aStream in the Reg. Ex. format: -*D:HH:MM:SS(.SS*)?."

	| d aHMS negative newDuration pos |
	#DtTmX.	"2001/07/18 Harmon, R. Added for <DurationExtra factory>."
	aStream skipWhile: [:c | c isWhitespace].
	negative := aStream peekFor: $-.
	pos := aStream position.
	d := Integer readFrom: aStream radix: 10.
	pos + 1 = aStream position ifFalse: [self errorInvalidFormat].
	(aStream peekFor: $:) ifFalse: [self errorInvalidFormat].
	aHMS := self readHMSFrom: aStream secondsRequired: true.
	newDuration := Duration 
				days: d
				hours: (aHMS at: 1)
				minutes: (aHMS at: 2)
				seconds: (aHMS at: 3).
	negative ifFalse: [^newDuration].
	^newDuration negated!

readHMSFrom: aStream secondsRequired: secondsRequired 
	"Private - Answers hh, mm, and ss list read from aStream <gettableStream> in the Reg. Ex. format: hh:mm:ss(.ss*)?.  Signal an error if not in specified format."

	| hh mm ss pos |
	#DtTmX.	"2001/07/18 Harmon, R. Added for DateAndTimeExtra support."
	pos := aStream position.
	hh := Integer readFrom: aStream radix: 10.
	pos + 2 = aStream position ifFalse: [self errorInvalidFormat].
	(aStream peekFor: $:) ifFalse: [self errorInvalidFormat].
	pos := aStream position.
	mm := Integer readFrom: aStream radix: 10.
	pos + 2 = aStream position ifFalse: [self errorInvalidFormat].
	(aStream peekFor: $:) 
		ifFalse: 
			[secondsRequired 
				ifTrue: [self errorInvalidFormat]
				ifFalse: 
					[^Array 
						with: hh
						with: mm
						with: 0]].
	ss := Number readFrom: aStream.
	^Array 
		with: hh
		with: mm
		with: ss!

readOffsetFrom: aStream 
	"Answers a new instance offset of the receiver read from aStream in the Reg. Ex. format: [-+]HH:MM:SS(.SS*)?."

	| negative newDuration aHMS |
	#DtTmX.	"2001/07/18 Harmon, R. Added for <DurationExtra factory>."
	negative := aStream peekFor: $-.
	negative ifFalse: [aStream peekFor: $+].
	aHMS := self readHMSFrom: aStream secondsRequired: false.
	newDuration := Duration 
				days: 0
				hours: (aHMS at: 1)
				minutes: (aHMS at: 2)
				seconds: (aHMS at: 3).
	negative ifFalse: [^newDuration].
	^newDuration negated!

seconds: seconds 
	"Answer a <Duration> which is <number> seconds in length
	If seconds is negative, answer a <Duration> that is abs (seconds) less 
	than <Duration factory> #zero.
	(idb) All numbers converted to a Fraction, Floats are rounded to 6 decimal places."

	^self new setSeconds: seconds asRestrictedFraction!

zero
	"Answer a <Duration> representing a length of no time."

	^self seconds: 0! !
!Duration class categoriesFor: #days:!instance creation!public! !
!Duration class categoriesFor: #days:hours:minutes:seconds:!instance creation!public! !
!Duration class categoriesFor: #errorInvalidFormat!exceptions!public! !
!Duration class categoriesFor: #fromString:!instance creation!public! !
!Duration class categoriesFor: #hours:!instance creation!public! !
!Duration class categoriesFor: #minutes:!instance creation!public! !
!Duration class categoriesFor: #readFrom:!instance creation!public! !
!Duration class categoriesFor: #readHMSFrom:secondsRequired:!helpers!public! !
!Duration class categoriesFor: #readOffsetFrom:!instance creation!public! !
!Duration class categoriesFor: #seconds:!instance creation!public! !
!Duration class categoriesFor: #zero!instance creation!public! !

Duration class methodProtocol: #'Duration factory' attributes: #(#ansi #readOnly) selectors: #(#~~ #~= #= #== #class #copy #days:hours:minutes:seconds: #doesNotUnderstand: #error: #hash #identityHash #isKindOf: #isMemberOf: #isNil #notNil #perform: #perform:with: #perform:with:with: #perform:with:with:with: #perform:withArguments: #printOn: #printString #respondsTo: #seconds: #yourself #zero)!

DurationToText guid: (GUID fromString: '{B7C9EDE3-D5D2-494F-8569-58EC71E88B89}')!
DurationToText comment: 'See [DolphinImageFolder]/Idb/Documentation for details

(C) 2005 Ian Bartholomew
ian@idb.me.uk
Public Domain Freeware'!
!DurationToText categoriesForClass!IDB Goodies!MVP-Type Converters-Text! !
!DurationToText methodsFor!

defaultFormat
	^'d:h:m:s'!

format
	format isNil ifTrue: [^self defaultFormat].
	^format!

format: aStringOrNil 
	format := aStringOrNil!

formatDaysIn: aDuration onto: aStream using: aString 
	| string |
	string := aDuration days printString.
	string size < aString size 
		ifTrue: [string := ((String new: aString size - string size) atAllPut: $0) , string].
	aStream nextPutAll: string!

formatHoursIn: aDuration onto: aStream using: aString 
	| hours string |
	hours := aString first isUppercase 
				ifTrue: 
					[self assert: [aStream position = 0].
					aDuration days * 24 + aDuration hours]
				ifFalse: [aDuration hours].
	string := hours printString.
	string size < aString size 
		ifTrue: [string := ((String new: aString size - string size) atAllPut: $0) , string].
	aStream nextPutAll: string!

formatMinutesIn: aDuration onto: aStream using: aString 
	| minutes string |
	minutes := aString first isUppercase 
				ifTrue: 
					[self assert: [aStream position = 0].
					(aDuration days * 24 + aDuration hours) * 60 + aDuration minutes]
				ifFalse: [aDuration minutes].
	string := minutes printString.
	string size < aString size 
		ifTrue: [string := ((String new: aString size - string size) atAllPut: $0) , string].
	aStream nextPutAll: string!

formatSecondsIn: aDuration onto: aStream using: aString 
	| string |
	string := aDuration seconds asFloat printString.
	string size < aString size 
		ifTrue: [string := ((String new: aString size - string size) atAllPut: $0) , string].
	aStream nextPutAll: string!

formatSelectorFor: aString 
	^##((LookupTable new)
		at: $d put: #formatDaysIn:onto:using:;
		at: $h put: #formatHoursIn:onto:using:;
		at: $m put: #formatMinutesIn:onto:using:;
		at: $s put: #formatSecondsIn:onto:using:;
		yourself) at: aString trimBlanks first asLowercase ifAbsent: [#formatError]!

leftToRight: aDuration 
	| stream |
	stream := String writeStream.
	(self format subStrings: $:) do: 
			[:each | 
			| formatString |
			formatString := each trimBlanks.
			self 
				perform: (self formatSelectorFor: formatString)
				with: aDuration
				with: stream
				with: formatString]
		separatedBy: [stream nextPut: $:].
	^stream contents!

rightToLeft: aString 
	"Convert aString into a Duration.  
		Expected format is day:hour:minute:seconds
			- any argument can have a sign
			- days hours and minutes are truncated to integers, seconds are not
			- arguments can be omitted from the left so 12:34 iseqivalent to 0:0:12:34
			- 0, more than 4 values , illegal characters answers nil"

	| values |
	(aString allSatisfy: [:each | '0123456789+-:.' includes: each]) ifFalse: [^self leftNullValue].
	values := OrderedCollection new.
	(aString subStrings: $:) reverseDo: [:each | values add: (Number fromString: each)].
	values isEmpty | (values size > 4) ifTrue: [^self leftNullValue].
	[values size < 4] whileTrue: [values add: 0].
	^Duration 
		days: (values at: 4) truncated
		hours: (values at: 3) truncated
		minutes: (values at: 2) truncated
		seconds: (values at: 1)! !
!DurationToText categoriesFor: #defaultFormat!accessing!public! !
!DurationToText categoriesFor: #format!accessing!public! !
!DurationToText categoriesFor: #format:!accessing!public! !
!DurationToText categoriesFor: #formatDaysIn:onto:using:!operations!public! !
!DurationToText categoriesFor: #formatHoursIn:onto:using:!operations!public! !
!DurationToText categoriesFor: #formatMinutesIn:onto:using:!operations!public! !
!DurationToText categoriesFor: #formatSecondsIn:onto:using:!operations!public! !
!DurationToText categoriesFor: #formatSelectorFor:!operations!public! !
!DurationToText categoriesFor: #leftToRight:!operations!public! !
!DurationToText categoriesFor: #rightToLeft:!operations!public! !

"Binary Globals"!

